#----------------------------------------------------------------------
#  Direct MDLSM method test - 3d heat sink
#  Same test case of PrePoMax tutorial #21
#  Author: Andrea Pavan
#  Date: 06/02/2023
#  License: GPLv3-or-later
#----------------------------------------------------------------------
using ElasticArrays;
using LinearAlgebra;
using SparseArrays;
using PyPlot;
include("utils.jl");


#problem definition
l1 = 54.4e-3;       #domain x size
l2 = 43e-3;     #domain y size
l3 = 60e-3;     #domain z size
tb = 3e-3;		#bottom thickness
tf = 4e-3;		#fins thickness
sf = 12.8e-3;		#fins spacing
k = 237;        #thermal conductivity
qb = 18382.4;       #bottom heat flux
Tamb = 20;      #ambient temperature
h = 23;     #convective film coefficient

meshSize = 1.5e-3;
#meshSize = 1.0e-3;
#meshSize = 0.75e-3;
surfaceMeshSize = meshSize;
minNeighbors = 60;
minSearchRadius = meshSize/2;


#read pointcloud from a SU2 file
time1 = time();
pointcloud = ElasticArray{Float64}(undef,3,0);      #2xN matrix containing the coordinates [X;Y;Z] of each node
boundaryNodes = Vector{Int}(undef,0);       #indices of the boundary nodes
internalNodes = Vector{Int}(undef,0);       #indices of the internal nodes
normals = ElasticArray{Float64}(undef,3,0);     #3xN matrix containing the components [nx;ny;nz] of the normal of each boundary node

pointcloud = parseSU2mesh("14_3d_heat_sink_mesh_32082.su2");
#pointcloud = parseSU2mesh("14_3d_heat_sink_mesh_83267.su2");
#pointcloud = parseSU2mesh("14_3d_heat_sink_mesh_169186.su2");
N = size(pointcloud,2);
for i=1:N
    if pointcloud[2,i]<=0+1e-5
        #bottom surface
        push!(boundaryNodes, i);
        append!(normals, [0,-1,0]);
    elseif l1-tf-1e-5<=pointcloud[1,i]<=l1+1e-5 && l2-1e-5<=pointcloud[2,i]<=l2+1e-5
        #upper surface (fin #4)
        push!(boundaryNodes, i);
        append!(normals, [0,1,0]);
    elseif l1-2*tf-sf-1e-5<=pointcloud[1,i]<=l1-tf-sf+1e-5 && l2-1e-5<=pointcloud[2,i]<=l2+1e-5
        #upper surface (fin #3)
        push!(boundaryNodes, i);
        append!(normals, [0,1,0]);
    elseif l1-3*tf-2*sf-1e-5<=pointcloud[1,i]<=l1-2*tf-2*sf+1e-5 && l2-1e-5<=pointcloud[2,i]<=l2+1e-5
        #upper surface (fin #2)
        push!(boundaryNodes, i);
        append!(normals, [0,1,0]);
    elseif 0-1e-5<=pointcloud[1,i]<=l1-3*tf-3*sf+1e-5 && l2-1e-5<=pointcloud[2,i]<=l2+1e-5
        #upper surface (fin #1)
        push!(boundaryNodes, i);
        append!(normals, [0,1,0]);
    elseif l1-tf-sf-1e-5<=pointcloud[1,i]<=l1-tf+1e-5 && 0+tb-1e-5<=pointcloud[2,i]<=0+tb+1e-5
        #upper surface (between fin #3 and fin #4)
        push!(boundaryNodes, i);
        append!(normals, [0,1,0]);
    elseif l1-2*tf-2*sf-1e-5<=pointcloud[1,i]<=l1-2*tf-sf+1e-5 && 0+tb-1e-5<=pointcloud[2,i]<=0+tb+1e-5
        #upper surface (between fin #2 and fin #3)
        push!(boundaryNodes, i);
        append!(normals, [0,1,0]);
    elseif l1-3*tf-3*sf-1e-5<=pointcloud[1,i]<=l1-3*tf-2*sf+1e-5 && 0+tb-1e-5<=pointcloud[2,i]<=0+tb+1e-5
        #upper surface (between fin #1 and fin #2)
        push!(boundaryNodes, i);
        append!(normals, [0,1,0]);
    elseif l1-1e-5<=pointcloud[1,i]<=l1+1e-5
        #right surface
        push!(boundaryNodes, i);
        append!(normals, [1,0,0]);
    elseif l1-tf-1e-5<=pointcloud[1,i]<=l1-tf+1e-5 && 0+tb-1e-5<=pointcloud[2,i]<=l2+1e-5
        #left surface (fin #4)
        push!(boundaryNodes, i);
        append!(normals, [-1,0,0]);
    elseif l1-tf-sf-1e-5<=pointcloud[1,i]<=l1-tf-sf+1e-5 && 0+tb-1e-5<=pointcloud[2,i]<=l2+1e-5
        #right surface (fin #3)
        push!(boundaryNodes, i);
        append!(normals, [1,0,0]);
    elseif l1-2*tf-sf-1e-5<=pointcloud[1,i]<=l1-2*tf-sf+1e-5 && 0+tb-1e-5<=pointcloud[2,i]<=l2+1e-5
        #left surface (fin #3)
        push!(boundaryNodes, i);
        append!(normals, [-1,0,0]);
    elseif l1-2*tf-2*sf-1e-5<=pointcloud[1,i]<=l1-2*tf-2*sf+1e-5 && 0+tb-1e-5<=pointcloud[2,i]<=l2+1e-5
        #right surface (fin #2)
        push!(boundaryNodes, i);
        append!(normals, [1,0,0]);
    elseif l1-3*tf-2*sf-1e-5<=pointcloud[1,i]<=l1-3*tf-2*sf+1e-5 && 0+tb-1e-5<=pointcloud[2,i]<=l2+1e-5
        #left surface (fin #2)
        push!(boundaryNodes, i);
        append!(normals, [-1,0,0]);
    elseif l1-3*tf-3*sf-1e-5<=pointcloud[1,i]<=l1-3*tf-3*sf+1e-5 && 0+tb-1e-5<=pointcloud[2,i]<=l2+1e-5
        #right surface (fin #1)
        push!(boundaryNodes, i);
        append!(normals, [1,0,0]);
    elseif 0-1e-5<=pointcloud[1,i]<=0+1e-5 && 0-1e-5<=pointcloud[2,i]<=l2+1e-5
        #left surface
        push!(boundaryNodes, i);
        append!(normals, [-1,0,0]);
    elseif 0-1e-5<=pointcloud[3,i]<=0+1e-5
        push!(boundaryNodes, i);
        append!(normals, [0,0,-1]);
    elseif l3-1e-5<=pointcloud[3,i]<=l3+1e-5
        push!(boundaryNodes, i);
        append!(normals, [0,0,1]);
    else
        push!(internalNodes, i);
        append!(normals, [0,0,0]);
    end
end

println("Generated pointcloud in ", round(time()-time1,digits=2), " s");
println("Pointcloud properties:");
println("  Boundary nodes: ",length(boundaryNodes));
println("  Internal nodes: ",length(internalNodes));
println("  Memory: ",memoryUsage(pointcloud,boundaryNodes));


#boundary conditions
N = size(pointcloud,2);     #number of nodes
g1 = zeros(Float64,N);
g2 = zeros(Float64,N);
g3 = zeros(Float64,N);
for i in boundaryNodes
    if pointcloud[2,i]<=0+1e-5
        #bottom surface
        g1[i] = 0.0;
        g2[i] = k;
        g3[i] = qb;
    else
        #everywhere else
        g1[i] = h;
        g2[i] = k;
        g3[i] = h*Tamb;
    end
end


#neighbor search (cartesian cells)
time2 = time();
neighbors = Vector{Vector{Int}}(undef,N);       #vector containing N vectors of the indices of each node neighbors
Nneighbors = zeros(Int,N);      #number of neighbors of each node
(neighbors,Nneighbors,cell) = cartesianNeighborSearch(pointcloud,meshSize,minNeighbors);

println("Found neighbors in ", round(time()-time2,digits=2), " s");
println("Connectivity properties:");
println("  Max neighbors: ",maximum(Nneighbors)," (at index ",findfirst(isequal(maximum(Nneighbors)),Nneighbors),")");
println("  Avg neighbors: ",round(sum(Nneighbors)/length(Nneighbors),digits=2));
println("  Min neighbors: ",minimum(Nneighbors)," (at index ",findfirst(isequal(minimum(Nneighbors)),Nneighbors),")");


#neighbors distances and weights
time3 = time();
P = Vector{Array{Float64}}(undef,N);        #relative positions of the neighbors
r2 = Vector{Vector{Float64}}(undef,N);      #relative distances of the neighbors
w = Vector{Vector{Float64}}(undef,N);       #neighbors weights
for i=1:N
    P[i] = Array{Float64}(undef,3,Nneighbors[i]);
    r2[i] = Vector{Float64}(undef,Nneighbors[i]);
    w[i] = Vector{Float64}(undef,Nneighbors[i]);
    for j=1:Nneighbors[i]
        P[i][:,j] = pointcloud[:,neighbors[i][j]]-pointcloud[:,i];
        r2[i][j] = P[i][:,j]'P[i][:,j];
    end
    r2max = maximum(r2[i]);
    for j=1:Nneighbors[i]
        w[i][j] = exp(-6*r2[i][j]/r2max);
        #w[i][j] = 1.0;
    end
end
wpde = 2.0;       #least squares weight for the pde
wbc = 2.0;        #least squares weight for the boundary condition


#least square matrix inversion
C = Vector{Matrix}(undef,N);        #stencil coefficients matrices
for i in internalNodes
    xj = P[i][1,:];
    yj = P[i][2,:];
    zj = P[i][3,:];
    V = zeros(Float64,4+4*Nneighbors[i],16);
    for j=1:Nneighbors[i]
        V[j,:] = [1, xj[j], yj[j], zj[j],  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0];
        V[j+Nneighbors[i],:] = [0, 0, 0, 0,  1, xj[j], yj[j], zj[j],  0, 0, 0, 0,  0, 0, 0, 0];
        V[j+2*Nneighbors[i],:] = [0, 0, 0, 0,  0, 0, 0, 0,  1, xj[j], yj[j], zj[j],  0, 0, 0, 0];
        V[j+3*Nneighbors[i],:] = [0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  1, xj[j], yj[j], zj[j]];
    end
    V[1+4*Nneighbors[i],:] = [0, 0, 0, 0,  0, 1, 0, 0,  0, 0, 1, 0,  0, 0, 0, 1];
    V[2+4*Nneighbors[i],:] = [0, k, 0, 0,  1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0];
    V[3+4*Nneighbors[i],:] = [0, 0, k, 0,  0, 0, 0, 0,  1, 0, 0, 0,  0, 0, 0, 0];
    V[4+4*Nneighbors[i],:] = [0, 0, 0, k,  0, 0, 0, 0,  0, 0, 0, 0,  1, 0, 0, 0];
    W = Diagonal(vcat(w[i],w[i],w[i],w[i],wpde,wpde,wpde,wpde));
    (Q,R) = qr(W*V);
    C[i] = inv(R)*transpose(Matrix(Q))*W;
end
for i in boundaryNodes
    xj = P[i][1,:];
    yj = P[i][2,:];
    zj = P[i][3,:];
    V = zeros(Float64,5+4*Nneighbors[i],16);
    for j=1:Nneighbors[i]
        V[j,:] = [1, xj[j], yj[j], zj[j],  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0];
        V[j+Nneighbors[i],:] = [0, 0, 0, 0,  1, xj[j], yj[j], zj[j],  0, 0, 0, 0,  0, 0, 0, 0];
        V[j+2*Nneighbors[i],:] = [0, 0, 0, 0,  0, 0, 0, 0,  1, xj[j], yj[j], zj[j],  0, 0, 0, 0];
        V[j+3*Nneighbors[i],:] = [0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  1, xj[j], yj[j], zj[j]];
    end
    V[1+4*Nneighbors[i],:] = [0, 0, 0, 0,  0, 1, 0, 0,  0, 0, 1, 0,  0, 0, 0, 1];
    V[2+4*Nneighbors[i],:] = [0, k, 0, 0,  1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0];
    V[3+4*Nneighbors[i],:] = [0, 0, k, 0,  0, 0, 0, 0,  1, 0, 0, 0,  0, 0, 0, 0];
    V[4+4*Nneighbors[i],:] = [0, 0, 0, k,  0, 0, 0, 0,  0, 0, 0, 0,  1, 0, 0, 0];
    V[5+4*Nneighbors[i],:] = [k*g1[i], 0, 0, 0,  -g2[i]*normals[1,i], 0, 0, 0,  -g2[i]*normals[2,i], 0, 0, 0,  -g2[i]*normals[3,i], 0, 0, 0];
    W = Diagonal(vcat(w[i],w[i],w[i],w[i],wpde,wpde,wpde,wpde,wbc));
    (Q,R) = qr(W*V);
    C[i] = inv(R)*transpose(Matrix(Q))*W;
end
println("Inverted least-squares matrices in ", round(time()-time3,digits=2), " s");


#matrix assembly
time4 = time();
rows = Int[];
cols = Int[];
vals = Float64[];
#rows = Vector{Int}(undef,4*N);
#cols = Vector{Int}(undef,4*N);
#vals = Vector{Float64}(undef,4*N);
for i=1:N
    #u equation
    push!(rows, i);
    push!(cols, i);
    push!(vals, 1);
    for j=1:Nneighbors[i]
        #u terms
        push!(rows, i);
        push!(cols, neighbors[i][j]);
        push!(vals, -C[i][1,j]);
        #qx terms
        push!(rows, i);
        push!(cols, N+neighbors[i][j]);
        push!(vals, -C[i][1,j+Nneighbors[i]]);
        #qy terms
        push!(rows, i);
        push!(cols, 2*N+neighbors[i][j]);
        push!(vals, -C[i][1,j+2*Nneighbors[i]]);
        #qz terms
        push!(rows, i);
        push!(cols, 3*N+neighbors[i][j]);
        push!(vals, -C[i][1,j+3*Nneighbors[i]]);
    end

    #qx equation
    push!(rows, i+N);
    push!(cols, i+N);
    push!(vals, 1);
    for j=1:Nneighbors[i]
        #u terms
        push!(rows, i+N);
        push!(cols, neighbors[i][j]);
        push!(vals, -C[i][5,j]);
        #qx terms
        push!(rows, i+N);
        push!(cols, N+neighbors[i][j]);
        push!(vals, -C[i][5,j+Nneighbors[i]]);
        #qy terms
        push!(rows, i+N);
        push!(cols, 2*N+neighbors[i][j]);
        push!(vals, -C[i][5,j+2*Nneighbors[i]]);
        #qz terms
        push!(rows, i+N);
        push!(cols, 3*N+neighbors[i][j]);
        push!(vals, -C[i][5,j+3*Nneighbors[i]]);
    end

    #qy equation
    push!(rows, i+2*N);
    push!(cols, i+2*N);
    push!(vals, 1);
    for j=1:Nneighbors[i]
        #u terms
        push!(rows, i+2*N);
        push!(cols, neighbors[i][j]);
        push!(vals, -C[i][9,j]);
        #qx terms
        push!(rows, i+2*N);
        push!(cols, N+neighbors[i][j]);
        push!(vals, -C[i][9,j+Nneighbors[i]]);
        #qy terms
        push!(rows, i+2*N);
        push!(cols, 2*N+neighbors[i][j]);
        push!(vals, -C[i][9,j+2*Nneighbors[i]]);
        #qz terms
        push!(rows, i+2*N);
        push!(cols, 3*N+neighbors[i][j]);
        push!(vals, -C[i][9,j+3*Nneighbors[i]]);
    end
    
    #qz equation
    push!(rows, i+3*N);
    push!(cols, i+3*N);
    push!(vals, 1);
    for j=1:Nneighbors[i]
        #u terms
        push!(rows, i+3*N);
        push!(cols, neighbors[i][j]);
        push!(vals, -C[i][13,j]);
        #qx terms
        push!(rows, i+3*N);
        push!(cols, N+neighbors[i][j]);
        push!(vals, -C[i][13,j+Nneighbors[i]]);
        #qy terms
        push!(rows, i+3*N);
        push!(cols, 2*N+neighbors[i][j]);
        push!(vals, -C[i][13,j+2*Nneighbors[i]]);
        #qz terms
        push!(rows, i+3*N);
        push!(cols, 3*N+neighbors[i][j]);
        push!(vals, -C[i][13,j+3*Nneighbors[i]]);
    end
end
M = sparse(rows,cols,vals,4*N,4*N);
println("Completed matrix assembly in ", round(time()-time4,digits=2), " s");


#linear system solution
time5 = time();
b = zeros(4*N);     #rhs vector
for i in internalNodes
    b[i] = 0 + 0 + 0 + 0;
    b[i+N] = 0 + 0 + 0 + 0;
    b[i+2*N] = 0 + 0 + 0 + 0;
    b[i+3*N] = 0 + 0 + 0 + 0;
end
for i in boundaryNodes
    b[i] = 0 + 0 + 0 + 0 + C[i][1,end]*k*g3[i];
    b[i+N] = 0 + 0 + 0 + 0 + C[i][5,end]*k*g3[i];
    b[i+2*N] = 0 + 0 + 0 + 0 + C[i][9,end]*k*g3[i];
    b[i+3*N] = 0 + 0 + 0 + 0 + C[i][13,end]*k*g3[i];
end
sol = M\b;
println("Linear system solved in ", round(time()-time5,digits=2), " s");
u = sol[1:N];
qx = sol[N+1:2*N];
qy = sol[2*N+1:3*N];
qz = sol[3*N+1:4*N];

#solution plot
figure();
plt = scatter3D(pointcloud[1,:],pointcloud[2,:],pointcloud[3,:],c=u,cmap="jet");
title("Numerical solution");
axis("equal");
colorbar(plt);
display(gcf());
